#lang racket/base

(require redex/reduction-semantics)
(provide (all-defined-out))

(define-language baseL
  [x ::= variable-not-otherwise-mentioned]
  [ι ::= natural]
  [op ::= * sub1 +])

(define (value? v)
  (or (exact-nonnegative-integer? v)
      (symbol? v)))

(define (op? v) (redex-match? baseL op v))

;; \lambda-calculus syntax
(define-extended-language lcL baseL
  [e ::= ι x (apply e e) (lambda (x) e) (let ([x e]) e)
     (ifz e e e)
     (op e ...)]
  #:binding-forms
  (lambda (x) e #:refers-to x)
  (let ([x e1]) e2 #:refers-to x))

;; evaluation for the \lambda-calculus
(define-extended-language eval-lcL lcL
  ; values
  [v ::= x ι (lambda (x) e)]
  ; strict evaluation contexts
  [E ::= hole (let ([x E]) e)
     (ifz E e e)
     (apply v ... E e ...)
     (op v ... E e ...)])

(define-metafunction baseL
  [(δ op any ...)
   ,(apply (eval (term op) (make-base-namespace)) (term (any ...)))])

(define intro-example
  '(+ (let ((x (apply f 5)))
        0)
      6))

;; Example from page 244/8 of Flanagan
(define running-example
  '(+ (+ 2 2) (let ([x 1]) (apply f x))))

;; Nested branching example
(define nested-branches
  `(let ([x (ifz (ifz (ifz 0 0 1)
                      0
                      1)
                 0
                 1)])
     LARGE))

;; fact 5
(define fact-example
  `(let ([Z (lambda (f)
              (let ([recursor (lambda (y)
                                (apply f (lambda (z) (apply (apply y y)
                                                            z))))])
                (apply recursor recursor)))])
     (let ([F (lambda (fact)
                (lambda (n)
                  (ifz n
                       1
                       (* n (apply fact (sub1 n))))))])
       (let ([fact (apply Z F)])
         (apply fact 5)))))
